/*
 * pipe.c (c) 2009-2015 JoergS
 *
 * functions for reading/writing compressed files or through pipes
 *    - MY_OPEN: bzopen, gzopen, popen, fopen, mmap, open
 *          
 *   ToDo: setvbuf could reduce syscalls to read (slow)
 *     but check that setvbuf works system dependend with strace (utils/vbuf.c)
 *     if so, we better should use primitiv read, write and own_cache (stdlib)
 *     but what about popen? (stdio) replace by pipe and fork/exec/system
 *       also read can be interupted (O_NONBLOCK, ret<len, select()),   
 *     popen may be faster on multiproc-PC because of async i/o and read before open (just after close),
 *      but check that theory by explicit timing/debugging incl. iostat
 *     may be popen should be replaced by own_mpi=(mpi or pipe or thread or seq)
 *     create MY_FILE for extern calls
 *     fwrite/write 2GB-limits?
 *
 * history:
 *  v0.5 - fork makes a (virtual) copy of the parent process (bigmem*2),
 *         which will be released by exec (ToDo: check with tracer!),
 *         this can be fail on some systems if memory < 2*used_mem
 *         solution(?): fork at before mallocs and fork in child on cmd ???
 *  v0.6 - popen(gzip) => same problem
 *         starting gzip (which needs only some kB) from spin (which
 *         has GB allocated) needs TEMPORARLY GB's for internal fork on IRIX64
 *         perror=Resource temporarily unavailable
 *  v0.7 - STOREH=2 one early fork, malloc, fork by child via commands like
 *  v0.9 - da STOREH=2 auf SGI-THOR bei >2GB >100h pg mit fehlermeldung stoppt
 *         STOREH=3 using zlib.h  (ca. 40% size, 30MB/s instead of 67MB/s)
 *         STOREH=4 using raw-files (save)
 *  v0.10 - STOREH 0 und 2 entfernt um einfacheren 
 *  v0.11 - do not print perror if errno==0
 *  v2.27 - memory only added STOREH=0, direct read/write STOREH=2
 * 
 * ToDo:
 *    async-IO: on disk with high BW one could do reading and processing
 *            at the same time
 *     but unfortunately there is no warranty that read_async returns 
 *     immediatly, it may block under certain conditions or bad implementations
 *     using my own (async)thread is also a problem, because
 *     it should share the cpu core with processing tasks (but not for gzip?)
 *     see speed_test.html where 650MB/s is consumed, so a async disk
 *     with same speed/thread will give SH-in-memory-performance (2013-03)
 * - readahead() blocks until the specified data has been read :(
 * - madvice MADV_SEQUENTIAL MADV_WILLNEED MADV_DONTDUMP (since Linux 3.4)
 *   mmap MAP_NORESERVE Do not reserve swap space
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  /* sleep() defined */
#include <string.h>
#include "error.h"
#include "pipe.h"

#ifdef _AIX
#include <sys/select.h>
#include <strings.h>
#endif

/* #include "config.h" */ /* is included by block.h */
#include "block.h"      /* B_NUM for 2*B_NUM async I/O buffers */


int pvv=0;	/* DEBUG-MODE parent-child-subchild communication */
const char *pipe_mode[5]={"realloc","pipe(gzip)","read","zlib","fread"};

/* v2.29: makes no sence because we store Hblocks already
 *        but Tru64+SunOS may read single bytes if unbuffered (use STOREH=2!)
 */
#ifndef PBUFL
#define PBUFL (1024*1024*8) /* 8MB(Hsize multiple), 0 = no buffering */
#endif

char *pipeHbuffer[B_NUM];  /* allocated outside in spins.c */
size_t pipeHlen[B_NUM];    /* set outside spins.c */
/* forking-methods dont work on IRIX with big calculations 
 *   because fork temporary doubles memory consumtion, 3 should work
 * 1=popen(gzip), 3=zlib_gzread 4=fread 0=mem 2=read
 */
#if   STOREH == 0
/* ---- use memory to get/store data ---- */ 
int      fn[B_NUM]; /* file-descriptor replacement == BlockNo */
/* 2015-04-08 Hbuffer is allocated as big area in spins.c before storeH and 
 *    not piecewise within storeH, avoiding OOM at mpich.xmalloc after
 *    old malloc(reserve),malloc(Hblock),free(reserve) which does not really
 *    free memory for further use (at least mpich's xmalloc fails)
 */
size_t pipeHused[B_NUM];
size_t pipeHpos[B_NUM];

/* 2015-03-09 malloc-functions are buggy, fall back to hfmax  */
extern ULong hfmax;  /* max. number of hxy-elements on disk  */

int mopen(int blk)  { pipeHpos[blk]=0; fn[blk]=blk; return(blk); }
int mclose(int blk) { pipeHpos[blk]=pipeHlen[blk];  return(blk); }

/* names mread/mwrite conflict with lam-mpi, renamed in 2.33 */
/* pointer is at last block, mixed r/w not allowed */
/* return len on success, return 0 on failure */
size_t pmwrite(void *from, size_t len, int blk) {
  if (!pipeHbuffer[blk]) return 0;
  if (pipeHlen[blk]<pipeHpos[blk]+len) {
    pipeHlen[blk]= pipeHpos[blk]; /* avoid later small package store */
    return 0; /* no space left */
  }
  memcpy(pipeHbuffer[blk]+pipeHpos[blk], from, len);
//  printf("\nDBG w %.0f %p",1.*len, pp);
  pipeHpos[blk]+=len;   /* go to the next block */
  if (pipeHused[blk] < pipeHpos[blk])
      pipeHused[blk] = pipeHpos[blk];
  return len;
}

/* faster direct access without copying, called by h_read() */
void *pmnext(int blk) {
  void *ptr; 
  if (!pipeHbuffer[blk]) {F0("noMem");return NULL;}
  if (pipeHpos[blk]+sizeof(struct thxy)>pipeHused[blk])
    {F0("notStored");return NULL;}
  ptr=pipeHbuffer[blk]+pipeHpos[blk];
  pipeHpos[blk]+=sizeof(struct thxy);
  return ptr;
}

size_t pmread(void *ptr, size_t len, int blk) {
//  printf("\nDBG r %.0f %.0f %p",1.*len,1.*mpos[blk]->len,mpos[blk]);
  if (!pipeHbuffer[blk]) {F0("noMem");return 0;}
  if (pipeHpos[blk]+len>pipeHused[blk]) {F0("notStored");return 0;}
  memcpy(ptr, pipeHbuffer[blk]+pipeHpos[blk], len);
  pipeHpos[blk]+=len;
  return len;
}

#define SET_BUF             /* no buffer used */
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    mopen(blk)
#define ROPEN(name,blk)    mopen(blk)
#define CLOSE   mclose
#define ZERO    -1
#define READ(buf,len,fn)   pmread( buf, len, fn)
#define WRITE(buf,len,fn)  pmwrite( buf, len, fn)
/* once per run */
#define INI(blk) {fn[blk]=ZERO;pipeHbuffer[blk]=NULL;pipeHlen[blk]=0;pipeHused[blk]=0;}

#elif   STOREH == 1
/* ---- use popen to read/write data ---- */
/* calls fork which can cause problems on systems doubling memory need */
/* #define FILETYPE FILE ToDo!, ToDo: | dd oflags=direct  */
FILE *fn[B_NUM];
#define SET_BUF
#define WOPENS "gzip -fcd %s"
#define ROPENS "gzip -fc1 >%s"
#define WOPEN(name,blk)    popen(name,"w")
#define ROPEN(name,blk)    popen(name,"r")
#define CLOSE   pclose
#define ZERO    (FILE *)0
#define READ(buf,len,fn)    fread( buf, 1, len, fn)
#define WRITE(buf,len,fn)  fwrite( buf, 1, len, fn)
#define INI(blk) fn[blk]=ZERO

#elif   STOREH == 2
/* ---- use read/write data ---- */ 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef O_NOATIME
#define O_NOATIME 0
#endif
#ifndef O_LARGEFILE    /* Tru64 */
#define O_LARGEFILE 0
#endif
#ifndef O_DIRECT
#define O_DIRECT        00040000        /* direct disk access hint */
#endif
int fn[B_NUM];  /* file handle */
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
/* +O_DIRECT 2015-04-09 avoid OOM and full (numa-)local memory
 *    PT+MPI_collect m streams to one interleaved stream for HDs 2013-04
 *  start with pt implementation and fix (1-2) files / node (local storage)
 *   later MPI? or mix speed + aio (test at pse.ovgu.de)
 *  Bsp: speed_mpi.gpl meggie hnz=35e6*35.38 HS=104MB/s(32Cores) Itspeed=23*SH
 * O_DIRECT needs _GNU_SOURCE defined, since 2.4.20 ... if defined()
 *   no caching (usually file does not fit into memory so cache wont help!)
 *   but caching eats memory and causes OOM on k2.6.x = bad kernels!
 *   for read it also needs aligned MEM-buffer (512..4k?)
 * O_NOATIME 2.6.8 not update inode on read
 *  readahead() is useless because it blocks, glibc-2.3 k2.4.13
 * posix_fadvise() POSIX_FADV_SEQUENTIAL POSIX_FADV_NOREUSE k2.5.60
 * aio_read: const blocklen? h_read(HSize=thxy=x*HBLen)
 */
#define WOPEN(name,blk) open(name,O_WRONLY|O_CREAT|O_LARGEFILE|O_DIRECT,6*64+4*8+4)
#define ROPEN(name,blk) open(name,O_RDONLY|O_LARGEFILE|O_NOATIME)
#define CLOSE   close
#define ZERO    0
#define READ(buf,len,fn)    read( fn, buf, len)
#define WRITE(buf,len,fn)  write( fn, buf, len)
#define INI(blk) fn[blk]=ZERO

#elif STOREH == 3  /* ToDo: pzip r/w one thread all for zip */
/* Jan2010: better good error message, than cryptic error later (if STOREH=3) */
#ifndef HAVE_ZLIB     
# error "no ZLIB available"
#endif
/* ---- use zlib or bzlib to read/write data ---- */
/* HAVE_ZLIB is defined in Makefile and given to gcc via -D macro=value */
#ifdef HAVE_BZLIB
  /* using bzlib (ca. 900kB of RAM used, bz2=1/3 unzip-time=4*gz) */
#include <bzlib.h>
BZFILE *fn[B_NUM];
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    BZ2_bzopen(name,"wb")
#define ROPEN(name,blk)    BZ2_bzopen(name,"rb")
#define CLOSE   BZ2_bzclose
#define ZERO    0
#define READ(buf,len,fn)   BZ2_bzread( fn, buf, len)
#define WRITE(buf,len,fn)  BZ2_bzwrite( fn, buf, len)
#define INI(blk) fn[blk]=ZERO
#else /* HAVE_BZLIB not defined */
  /* using zlib (ca. 260kB of dyn. RAM used, gz=1/2) */
#include <zlib.h>
gzFile fn[B_NUM];
#define SET_BUF
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    gzopen(name,"wb2f")
#define ROPEN(name,blk)    gzopen(name,"rb")
#define CLOSE   gzclose
#define ZERO    0
#define READ(buf,len,fn)   gzread( fn, buf, len)
#define WRITE(buf,len,fn)  gzwrite( fn, buf, len)
#define INI(blk) fn[blk]=ZERO
#endif /* !BZLIB */

#elif STOREH == 4	/* std-file */
/* ---- default method using fread/fwrite data ---- */
FILE *fn[B_NUM];
#if PBUFL>0
char pbuf[B_NUM][PBUFL];
#define SET_BUF  setvbuf(fn[block],pbuf[block],_IOFBF,PBUFL)
#else
#define SET_BUF  setvbuf(fn[block],(char *)0,_IONBF,0)
#endif
#define WOPENS "%s"
#define ROPENS "%s"
#define WOPEN(name,blk)    fopen(name,"wb")
#define ROPEN(name,blk)    fopen(name,"rb")
#define CLOSE   fclose
#define ZERO    (FILE *)0
#define READ(buf,len,fn)  (fread( buf, len, 1, fn)*len)
#define WRITE(buf,len,fn) (fwrite( buf, len, 1, fn)*len)
#define INI(blk) fn[blk]=ZERO
/* see utils/vbuf.c for i/o-experience */
#else
# error "no such STOREH defined"
#endif

/* ------------------------------- close ----------------------- */
int pipe_close(int block){
 if (fn[block]!=ZERO) CLOSE(fn[block]); /* only close if open */
 fn[block]=ZERO;                  /* mark as closed */
 return 0;
}
/* ------------------------------- read ------------------------ */
int pipe_ropen(char *name,int block){
    char s[100];int i; sprintf(s,WOPENS,name);
    if (block<0 || block>=B_NUM) return -1;  /* avoid gcc warnings */
    if (fn[block]!=ZERO) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=ZERO;i<5 && fn[block]==ZERO;i++){
      fn[block]=ROPEN(s,block);
      if (fn[block]==ZERO) { E2("pipe_ropen %s %d",s,i);perror(" "); 
       sleep(60); 
      }
    } if (fn[block]==ZERO)   E0("pipe_ropen failed");
    if( i>1 ) E0("pipe_ropen => ok");
    SET_BUF;
    return 0;
}
int pipe_read(char *buf,int len,int block){
    int nbytes=0; if(buf)buf[0]=0;
    if (fn[block]==ZERO) return 0;
    nbytes    = (int)READ( buf, len, fn[block]);
/*
    if (nbytes != len){ if(errno) perror("pipe_read");
      nbytes += (int)READ( buf+nbytes, len-nbytes, fn[block]);
      if (nbytes != len){ if(errno) perror("pipe_read"); }
    }
*/
    return nbytes;
}
/* --------------------------- write --------------------- */
int pipe_wopen(char *name,int block){
    char s[100];int i; sprintf(s,ROPENS,name);
    if (block<0 || block>=B_NUM) return -1;  /* avoid gcc warnings 2015-02 */
    if (fn[block]!=ZERO) pipe_close(block); /* forgotten close? */
    for(i=0,fn[block]=ZERO;i<5 && fn[block]==ZERO;i++){
      fn[block]=WOPEN(s,block);
      if (fn[block]==ZERO) { E2("pipe_wopen %s %d",s,i);perror(" ");
       sleep(30); }
    } 
    if (fn[block]==ZERO) { E0("pipe_wopen => failed"); return 1; }
    if( i>1 ) E0("pipe_wopen => ok");
    SET_BUF;
    return 0;
}
int pipe_write(char *s,int l,int block){
    int nbytes;
    if (fn[block]==ZERO) return 0;
    nbytes=(int)WRITE(s,l,fn[block]);
    /* reduce output to log mpi_n, mostly its on first or all nodes  */
    /* PERROR outside pipe.c */
    /* if(nbytes != l && (mpi_i&(mpi_i-1))==0) perror("pipe_write"); */
    return nbytes;
}
int sub_ini(){
  int i;
  if (!mpi_i)
    printf(" pipe.c sub_ini STOREH=%d %s 8*sizeof(size_t)=%d\n",
            STOREH, pipe_mode[STOREH], 8*(int)sizeof(size_t));
  for(i=0;i<B_NUM;i++) INI(i);
  return 0;
}
int sub_end(){return 0;}
/* ================================================================= */

#if 0		/* test this */
int buf[1024*1024];
int main()
{       int i,j,k;char s[100+1];
	FILE *f1;
	setvbuf(stdout,(char *)NULL,_IONBF,0);
        sub_ini();				/* do this early!!! */
        s[100]=0;buf[0]=buf[1024*1023]=1;
#define FLEN 100 /* 100x18 */
        pipe_wopen("tmp000.gz");			/* gzip */
        for(i=1;i<FLEN;i++)
        pipe_write("Das ist ein Test.\n",18);
        pipe_close();
        printf("tmp000.gz written\n"); sleep(10);
        for(k=0;k<100000;k++){
          pipe_ropen("tmp000.gz");			/* gzip */
          for(i=1;i<FLEN;i++){
            j=pipe_read(s,18);s[j]=0;		/* timeout-zeit? */
            if(j!=18) fprintf(stderr," error pipe_read(%d)=%d\n",k,j);
            printf("%s",s);
          }
          pipe_close();
        }
        sub_end();
	return 0;
}
#endif
